/*
 * OpenPBX -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2005, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * See http://www.openpbx.org for more information about
 * the OpenPBX project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Playback a file with audio detect
 * 
 */
#ifdef HAVE_CONFIG_H
#include "confdefs.h"
#endif
 
#include <string.h>
#include <stdlib.h>

#include "openpbx.h"

OPENPBX_FILE_VERSION("$HeadURL: svn://svn.openpbx.org/openpbx/trunk/apps/app_talkdetect.c $", "$Revision: 1055 $")

#include "openpbx/lock.h"
#include "openpbx/file.h"
#include "openpbx/logger.h"
#include "openpbx/channel.h"
#include "openpbx/pbx.h"
#include "openpbx/module.h"
#include "openpbx/translate.h"
#include "openpbx/utils.h"
#include "openpbx/dsp.h"

static char *tdesc = "Playback with Talk Detection";

static char *app = "BackgroundDetect";

static char *synopsis = "Background a file with talk detect";

static char *descrip = 
"  BackgroundDetect(filename[|sil[|min|[max]]]):  Plays  back  a  given\n"
"filename, waiting for interruption from a given digit (the digit must\n"
"start the beginning of a valid extension, or it will be ignored).\n"
"During the playback of the file, audio is monitored in the receive\n"
"direction, and if a period of non-silence which is greater than 'min' ms\n"
"yet less than 'max' ms is followed by silence for at least 'sil' ms then\n"
"the audio playback is aborted and processing jumps to the 'talk' extension\n"
"if available.  If unspecified, sil, min, and max default to 1000, 100, and\n"
"infinity respectively.  Returns -1 on hangup, and 0 on successful playback\n"
"completion with no exit conditions.\n";

STANDARD_LOCAL_USER;

LOCAL_USER_DECL;

static int background_detect_exec(struct opbx_channel *chan, void *data)
{
	int res = 0;
	struct localuser *u;
	char *tmp;
	char *options;
	char *stringp;
	struct opbx_frame *fr;
	int notsilent=0;
	struct timeval start = { 0, 0};
	int sil = 1000;
	int min = 100;
	int max = -1;
	int x;
	int origrformat=0;
	struct opbx_dsp *dsp;
	
	if (opbx_strlen_zero(data)) {
		opbx_log(LOG_WARNING, "BackgroundDetect requires an argument (filename)\n");
		return -1;
	}

	LOCAL_USER_ADD(u);

	tmp = opbx_strdupa(data);
	if (!tmp) {
		opbx_log(LOG_ERROR, "Out of memory\n");
		LOCAL_USER_REMOVE(u);
		return -1;
	}	

	stringp=tmp;
	strsep(&stringp, "|");
	options = strsep(&stringp, "|");
	if (options) {
		if ((sscanf(options, "%d", &x) == 1) && (x > 0))
			sil = x;
		options = strsep(&stringp, "|");
		if (options) {
			if ((sscanf(options, "%d", &x) == 1) && (x > 0))
				min = x;
			options = strsep(&stringp, "|");
			if (options) {
				if ((sscanf(options, "%d", &x) == 1) && (x > 0))
					max = x;
			}
		}
	}
	opbx_log(LOG_DEBUG, "Preparing detect of '%s', sil=%d,min=%d,max=%d\n", 
						tmp, sil, min, max);
	if (chan->_state != OPBX_STATE_UP) {
		/* Otherwise answer unless we're supposed to send this while on-hook */
		res = opbx_answer(chan);
	}
	if (!res) {
		origrformat = chan->readformat;
		if ((res = opbx_set_read_format(chan, OPBX_FORMAT_SLINEAR))) 
			opbx_log(LOG_WARNING, "Unable to set read format to linear!\n");
	}
	if (!(dsp = opbx_dsp_new())) {
		opbx_log(LOG_WARNING, "Unable to allocate DSP!\n");
		res = -1;
	}
	if (!res) {
		opbx_stopstream(chan);
		res = opbx_streamfile(chan, tmp, chan->language);
		if (!res) {
			while(chan->stream) {
				res = opbx_sched_wait(chan->sched);
				if (res < 0) {
					res = 0;
					break;
				}
				res = opbx_waitfor(chan, res);
				if (res < 0) {
					opbx_log(LOG_WARNING, "Waitfor failed on %s\n", chan->name);
					break;
				} else if (res > 0) {
					fr = opbx_read(chan);
					if (!fr) {
						res = -1;
						break;
					} else if (fr->frametype == OPBX_FRAME_DTMF) {
						char t[2];
						t[0] = fr->subclass;
						t[1] = '\0';
						if (opbx_canmatch_extension(chan, chan->context, t, 1, chan->cid.cid_num)) {
							/* They entered a valid  extension, or might be anyhow */
							res = fr->subclass;
							opbx_frfree(fr);
							break;
						}
					} else if ((fr->frametype == OPBX_FRAME_VOICE) && (fr->subclass == OPBX_FORMAT_SLINEAR)) {
						int totalsilence;
						int ms;
						res = opbx_dsp_silence(dsp, fr, &totalsilence);
						if (res && (totalsilence > sil)) {
							/* We've been quiet a little while */
							if (notsilent) {
								/* We had heard some talking */
								ms = opbx_tvdiff_ms(opbx_tvnow(), start);
								ms -= sil;
								if (ms < 0)
									ms = 0;
								if ((ms > min) && ((max < 0) || (ms < max))) {
									char ms_str[10];
									opbx_log(LOG_DEBUG, "Found qualified token of %d ms\n", ms);

									/* Save detected talk time (in milliseconds) */ 
									sprintf(ms_str, "%d", ms );	
									pbx_builtin_setvar_helper(chan, "TALK_DETECTED", ms_str);
									
									opbx_goto_if_exists(chan, chan->context, "talk", 1);
									res = 0;
									opbx_frfree(fr);
									break;
								} else
									opbx_log(LOG_DEBUG, "Found unqualified token of %d ms\n", ms);
								notsilent = 0;
							}
						} else {
							if (!notsilent) {
								/* Heard some audio, mark the begining of the token */
								start = opbx_tvnow();
								opbx_log(LOG_DEBUG, "Start of voice token!\n");
								notsilent = 1;
							}
						}
						
					}
					opbx_frfree(fr);
				}
			}
			opbx_stopstream(chan);
		} else {
			opbx_log(LOG_WARNING, "opbx_streamfile failed on %s for %s\n", chan->name, (char *)data);
			res = 0;
		}
	}
	if (res > -1) {
		if (origrformat && opbx_set_read_format(chan, origrformat)) {
			opbx_log(LOG_WARNING, "Failed to restore read format for %s to %s\n", 
				chan->name, opbx_getformatname(origrformat));
		}
	}
	if (dsp)
		opbx_dsp_free(dsp);
	LOCAL_USER_REMOVE(u);
	return res;
}

int unload_module(void)
{
	STANDARD_HANGUP_LOCALUSERS;
	return opbx_unregister_application(app);
}

int load_module(void)
{
	return opbx_register_application(app, background_detect_exec, synopsis, descrip);
}

char *description(void)
{
	return tdesc;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}


